--== [ VARIABLES ] ==--
--[ Global variables for RecipeBook ]--
RecipeBookData = {};
--[ Options data ]--
local Realm, Faction, OtherFaction, Player;
-- Monitoring status of game tooltip/chatframe
local RB_TooltipUpdated, RB_ChatFrameUpdated = false, false;
--[ Debug ]--
local RB_Debug = false;
local RB_Initialized = false;

--[ Misc ]--
local RB_TradeskillData = {};
local RB_LastWhat = {nil, false, {}};
local RB_SessionActive = false; --{session status, who, whatskill}
local RB_SessionQueue = {}; --for /rb send all

--[ Hooked Function Variables ]--
local RecipeBook_Old_GameTooltip_OnHide;
local RecipeBook_Old_ContainerFrameItemButton_OnEnter;
local RecipeBook_Old_ContainerFrameItemButton_OnUpdate;
local RecipeBook_Old_AuctionFrameItem_OnEnter;
local RecipeBook_Old_SetItemRef;
local RecipeBook_Old_GameTooltip_SetInventoryItem;
local RecipeBook_Old_GameTooltip_SetMerchantItem;
local RecipeBook_Old_ChatFrame_OnEvent;


--== [ ON LOAD ] ==--
function RecipeBook_OnLoad()
	--== Configuration Frame ==--
	UIPanelWindows["RecipeBook_OptionsFrame"] = {area = "center", pushable = 0};
	tinsert(UISpecialFrames, "RecipeBook_OptionsFrame");
	--== Event Registration ==--
	this:RegisterEvent("ADDON_LOADED");
	this:RegisterEvent("PLAYER_ENTERING_WORLD"); --For loading data
	this:RegisterEvent("TRADE_SKILL_SHOW");
	this:RegisterEvent("TRADE_SKILL_UPDATE");
	this:RegisterEvent("CRAFT_SHOW");
	this:RegisterEvent("CRAFT_UPDATE"); --These four are the workhorse of the DB
	this:RegisterEvent("AUCTION_ITEM_LIST_UPDATE"); --Auction list updating can otherwise clear tooltips
	this:RegisterEvent("CHAT_MSG_SKILL"); --Skillups for disenchants make baby enchanters happy.
	this:RegisterEvent("CHAT_MSG_SYSTEM");
	this:RegisterEvent("CHAT_MSG_WHISPER");
	this:RegisterEvent("CHAT_MSG_WHISPER_INFORM");
	
	--Slash Command Handlers--
	SlashCmdList["RecipeBook"] = RecipeBook_SlashHandler;
	SLASH_RecipeBook1 = "/recipebook";
	SLASH_RecipeBook2 = "/rb";
	
	DEFAULT_CHAT_FRAME:AddMessage(RECIPEBOOK_LOADED); 

	--==  Function Hooks ==--
	-- Blizzard function hooks --
	RecipeBook_Old_AuctionFrameItem_OnEnter = AuctionFrameItem_OnEnter; --Auction Frame
	AuctionFrameItem_OnEnter = RecipeBook_AuctionFrameItem_OnEnter;

	RecipeBook_Old_ContainerFrameItemButton_OnEnter = ContainerFrameItemButton_OnEnter; --Good ol' bags
	ContainerFrameItemButton_OnEnter = RecipeBook_ContainerFrameItemButton_OnEnter;
	
	RecipeBook_Old_ContainerFrameItemButton_OnUpdate = ContainerFrameItemButton_OnUpdate; --Bags update!
	ContainerFrameItemButton_OnUpdate = RecipeBook_ContainerFrameItemButton_OnUpdate;

	RecipeBook_Old_GameTooltip_OnHide = GameTooltip_OnHide;  --Tidy up data
	GameTooltip_OnHide = RecipeBook_GameTooltip_OnHide;
	
	RecipeBook_Old_SetItemRef = SetItemRef; --Chat links
	SetItemRef = RecipeBook_SetItemRef;
	
	RecipeBook_Old_GameTooltip_SetLootItem = GameTooltip.SetLootItem; --Loot items
	GameTooltip.SetLootItem = RecipeBook_GameTooltip_SetLootItem;
	

	RecipeBook_Old_GameTooltip_SetInventoryItem = GameTooltip.SetInventoryItem; --Bank slots
	GameTooltip.SetInventoryItem = RecipeBook_GameTooltip_SetInventoryItem;
	
	RecipeBook_Old_GameTooltip_SetMerchantItem = GameTooltip.SetMerchantItem; --Bank slots
	GameTooltip.SetMerchantItem = RecipeBook_GameTooltip_SetMerchantItem;
	
	RecipeBook_Old_ChatFrame_OnEvent = ChatFrame_OnEvent;
	ChatFrame_OnEvent = RecipeBook_ChatFrame_OnEvent;

	
	--== Specific hooks for specific mods ==--
	-- My Inventory	--
	if(MyInventoryProfile) then 
		RecipeBook_Old_MIFrameItemButton_OnEnter = MyInventory_FrameItemButton_OnEnter;
		MyInventory_FrameItemButton_OnEnter = RecipeBook_MIFrameItemButton_OnEnter;
		RecipeBook_Old_MIFrameItemButton_OnUpdate = MyInventoryFrameItemButton_OnUpdate;
		MyInventoryFrameItemButton_OnUpdate = RecipeBook_MIFrameItemButton_OnUpdate;
	end
	-- AllInOne Inventory --
	if(AllInOneInventory_ReplaceBags or AllInOneInventory_ReplaceBags == 0) then 
		RecipeBook_Old_AIOI_ModifyItemTooltip = AllInOneInventory_ModifyItemTooltip;
		AllInOneInventory_ModifyItemTooltip = RecipeBook_AIOI_ModifyItemTooltip;
	end
	-- BankStatement --
	if(BankStatementItemButton_OnEnter) then --Checking for BankStatement to play nice.
		RecipeBook_Old_BS_ItemButton_OnEnter = BankStatementItemButton_OnEnter;
		BankStatementItemButton_OnEnter = RecipeBook_BS_ItemButton_OnEnter;
		
		RecipeBook_Old_BS_ContainerFrameItemButton_OnEnter = BankStatementContainerFrameItemButton_OnEnter;
		BankStatementContainerFrameItemButton_OnEnter = RecipeBook_BS_ContainerFrameItemButton_OnEnter;
	end
	-- BankItems --
	if(BankItems_Button_OnEnter) then 
		RecipeBook_Old_BI_Button_OnEnter = BankItems_Button_OnEnter;
		BankItems_Button_OnEnter = RecipeBook_BI_Button_OnEnter;
		
		RecipeBook_Old_BI_BagItem_OnEnter = BankItems_BagItem_OnEnter;
		BankItems_BagItem_OnEnter = RecipeBook_BI_BagItem_OnEnter;
	end
	
	--== Static Popup Windows for data send ==--
	StaticPopupDialogs["RECIPEBOOK_REQUESTING_SEND"] = {
		text = " ",
		button1 = TEXT(ACCEPT),
		button2 = TEXT(DECLINE),
		OnShow = function()
			this.timeleft = RECIPEBOOK_SEND_TIMEOUT;
		end,
		OnAccept = function()
			RecipeBook_Debug("Accepted Session.");
			RB_Messenger_AcceptSession(RB_Messenger_GetName(), RB_Messenger_GetTradeskill());
		end,
		OnCancel = function ()
			RecipeBook_Debug("Cancelled Session.");
			RB_Messenger_CancelSession(RB_Messenger_GetName(), RB_Messenger_GetTradeskill());
		end,
		OnUpdate = function(elapsed)
			local text = getglobal(this:GetName().."Text");
			local timeleft = ceil(this.timeleft);
			if ( timeleft < 60 ) then
				text:SetText(format(RECIPEBOOK_POPUP_RECEIVE, RB_Messenger_GetName(), RB_Messenger_GetTradeskill(), timeleft, GetText("SECONDS", nil, timeleft)));
			else
				text:SetText(format(RECIPEBOOK_POPUP_RECEIVE, RB_Messenger_GetName(), RB_Messenger_GetTradeskill(), ceil(timeleft / 60), GetText("MINUTES", nil, ceil(timeleft / 60))));
			end
			StaticPopup_Resize(this, "RECIPEBOOK_REQUESTING_SEND");
		end,
		timeout = RECIPEBOOK_SEND_TIMEOUT,
		whileDead = 1;
	};
		StaticPopupDialogs["RECIPEBOOK_AWAITING_SEND"] = {
		text = " ",
		button1 = TEXT(CANCEL),
		OnShow = function()
			this.timeleft = RECIPEBOOK_SEND_TIMEOUT;
		end,
		OnAccept = function ()
			RB_Messenger_CancelSession(RB_Messenger_GetName(), RB_Messenger_GetTradeskill());
		end,
		OnUpdate = function(elapsed)
			local text = getglobal(this:GetName().."Text");
			local timeleft = ceil(this.timeleft);
			if ( timeleft < 60 ) then
				text:SetText(format(RECIPEBOOK_POPUP_SEND,  RB_Messenger_GetTradeskill(), RB_Messenger_GetName(), timeleft, GetText("SECONDS", nil, timeleft)));
			else
				text:SetText(format(RECIPEBOOK_POPUP_SEND, RB_Messenger_GetTradeskill(), RB_Messenger_GetName(), ceil(timeleft / 60), GetText("MINUTES", nil, ceil(timeleft / 60))));
			end
			StaticPopup_Resize(this, "RECIPEBOOK_AWAITING_SEND");
		end,
		timeout = RECIPEBOOK_SEND_TIMEOUT,
		whileDead = 1;
	};
	StaticPopupDialogs["RECIPEBOOK_AUTO_SEND"] = {
		text = " ",
		button1 = TEXT(CANCEL),
		OnShow = function()
			this.timeleft = RECIPEBOOK_AUTOSEND_TIMEOUT;
		end,
		OnAccept = function ()
			RB_Messenger_CancelSession(RB_Messenger_GetName(), RB_Messenger_GetTradeskill());
		end,
		OnCancel = function ()
			RecipeBook_Debug("Completed Timeout.");
			RB_Messenger_InitiateSession(RB_Messenger_GetName(), "Queue");
		end,
		OnUpdate = function(elapsed)
			local text = getglobal(this:GetName().."Text");
			local timeleft = ceil(this.timeleft);
			if ( timeleft < 60 ) then
				text:SetText(format(RECIPEBOOK_POPUP_PAUSE, timeleft, GetText("SECONDS", nil, timeleft)));
			else
				text:SetText(format(RECIPEBOOK_POPUP_PAUSE, ceil(timeleft / 60), GetText("MINUTES", nil, ceil(timeleft / 60))));
			end
			StaticPopup_Resize(this, "RECIPEBOOK_AUTO_SEND");
		end,
		timeout = RECIPEBOOK_AUTOSEND_TIMEOUT,
		whileDead = 1;
	};
end


--== [ INITIALIZE ] ==--
-- [ InitializeSetup() : New Player? New Data? This makes sure all the options are in the right place. ] --
local function RecipeBook_InitializeSetup()
	RecipeBook_Debug("InitializeSetup Called.");
	if Faction == nil then 
		RecipeBook_SetGlobals();
		if Faction == nil then 
			RecipeBook_Print("No Faction data available for this character.  Defaulting to Alliance.  PLEASE contact Ayradyss (dr.nykki@gmail.com) and inform her.", 1);
			Faction = FACTION_ALLIANCE;
		end
	end	

	if RecipeBookData == nil then 	--No RecipeBook Data at all.
		RecipeBook_Debug("Nil data");
		RecipeBookData = {}; 
		RecipeBookData.Version = 2;
	end
	if RecipeBookData.Version == nil then -- If you're older than 0.5 then you've been playing with a nonfunctional RecipeBook for several patches.
		RecipeBook_Debug("Nil version.");
		if RecipeBookData[Realm] ~= nil then RecipeBook_Print("Your RecipeBook is too outdated for conversion.  You're going to have to start over.", 1) end;
		RecipeBookData = {};
		RecipeBookData.Version = 2;
	end
	--Setting up data structures.
	if RecipeBookData[Realm] == nil then --This realm
		RecipeBookData[Realm] = {};
	end
	if RecipeBookData[Realm][Faction] == nil then --This faction (now localized!)
		RecipeBookData[Realm][Faction] = {};
	end
	if RecipeBookData[Realm][Faction]["Personal"] == nil then --Your characters
		RecipeBookData[Realm][Faction]["Personal"] = {};
	end
	if RecipeBookData[Realm][Faction]["Personal"][Player] == nil then --This character
		RecipeBookData[Realm][Faction]["Personal"][Player] = {};
	end
	if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] == nil then --This character's options (note: This is going to be moved to per-character/per-account savedvariables in 1.9, which is going to make a number of things much nicer.)
		RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
	end
	if RecipeBookData[Realm][Faction]["Shared"] == nil then --Any shared data.
		RecipeBookData[Realm][Faction]["Shared"] = {};
	end
	if RecipeBookData[Realm][Faction]["Banked"] == nil then --Any banked items (saved by faction, not by who banked 'em.)
		RecipeBookData[Realm][Faction]["Banked"] = {};
	end
	
	if tonumber(RecipeBookData.Version) < 2 then --Conversion to locale-specific data in data version 2; also some old options are converted.
		if (FACTION_ALLIANCE == "Alliance" and FACTION_HORDE == "Horde") then --No change.
		else
			local function movefaction(lrealm, data) --Copies existing data to locale-specifics.
				if lrealm == "Version" then return nil end;
				if RecipeBookData[lrealm]["Alliance"] ~= nil then
					RecipeBookData[lrealm][FACTION_ALLIANCE] = RecipeBookData[lrealm]["Alliance"];
					RecipeBookData[lrealm]["Alliance"] = nil;
				end
				if RecipeBookData[lrealm]["Horde"] ~= nil then
					RecipeBookData[lrealm][FACTION_HORDE] = RecipeBookData[lrealm]["Horde"];
					RecipeBookData[lrealm]["Horde"] = nil;
				end
			end
			table.foreach(RecipeBookData, movefaction)
		end
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] ~= nil then
			-- Output ChatFrame|Tooltip --> Chatframe/Tooltip On|Off
			if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] == "Tooltip" then 
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Tooltip"] = "On";
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] = nil;
			elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] == "ChatFrame" then 
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["ChatFrame"] = "On";
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["Output"] = nil;
			end
			-- AltData --> AltFaction, do this before SameFaction/OtherFaction conversion
			if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"] ~= nil then 
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] = RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"];
				RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltData"] = nil;
			end --Old data change.
			-- AltFaction Same|Opposite|Both --> SameFaction/OtherFaction On|Off
			if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] ~= nil then 
				if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Same" then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "On"
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "Off";
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] = nil;
				elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Opposite" then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "Off"
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "On";
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] = nil;
				elseif RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] == "Both" then 
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["SameFaction"] = "On"
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["OtherFaction"] = "On";
					RecipeBookData[Realm][Faction]["Personal"][Player]["Options"]["AltFaction"] = nil;
				end
			end
		else
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
		end
	end
	RecipeBookData.Version = 2; --Conversion should only run once.
	RB_Initialized = true;
end

local function RecipeBook_SetGlobals()
	Player = UnitName("player");
	Realm = GetRealmName();
	Faction = UnitFactionGroup("player");
	if Faction ~= nil then
		if FACTION_ALLIANCE ~= "Alliance" then --Blizzard's UnitFactionGroup does not return localized data.
			if (Faction == "Alliance" or Faction == FACTION_ALLIANCE) then Faction = FACTION_ALLIANCE;
			else Faction = FACTION_HORDE;
			end
		end
		OtherFaction = (Faction == FACTION_HORDE and FACTION_ALLIANCE or FACTION_HORDE);
	end
end


--== [ EVENT HANDLERS ] ==--
-- [ OnEvent(event) : So what happens when an event fires? ] --
function RecipeBook_OnEvent(event)
	RecipeBook_Debug(event);
	if (event == "PLAYER_ENTERING_WORLD") then
		if RB_Initialized then return end;
		RecipeBook_SetGlobals();
		RecipeBook_InitializeSetup();
	elseif(event == "CRAFT_SHOW" or event == "CRAFT_UPDATE") then
		RecipeBook_UpdateSkillInfo("craft"); --Enchanting window
	elseif(event=="TRADE_SKILL_SHOW" or event == "TRADE_SKILL_UPDATE") then
		RecipeBook_UpdateSkillInfo("skill"); --All other tradeskills.
	elseif(event=="AUCTION_ITEM_LIST_UPDATE") then
		RecipeBook_DoHookedFunction();
	elseif(event=="CHAT_MSG_SKILL") then
		RecipeBook_ParseSkillupMessage(arg1);
	elseif(event == "CHAT_MSG_SYSTEM") then
		RecipeBook_ParseSystemMessage(arg1);
	elseif(event == "ADDON_LOADED") then
		if string.lower(arg1) == "recipebook" then 
			RecipeBook_SetGlobals();
			if Faction ~= nil then 
				RecipeBook_InitializeSetup();
			end
		elseif arg1 == "Blizzard_AuctionUI" then 
			RecipeBook_Old_AuctionFrameItem_OnEnter = AuctionFrameItem_OnEnter; --Auction Frame
			AuctionFrameItem_OnEnter = RecipeBook_AuctionFrameItem_OnEnter;
		end
	end
	
end

--[ SlashHandler: for all those nifty /rb commands ]--
function RecipeBook_SlashHandler(msg)
	local msglist = {};
	local toggle;
	--/rb : Display status.
	if (not msg or msg == "") then 
		RecipeBook_PrintStatus("all");
		return;
	else
		msglist = RecipeBook_Split(string.lower(msg));
	end
	if not msglist[1] then
		RecipeBook_Print(RECIPEBOOK_USAGE, -1);
	--/rb config : Toggles the config window
	elseif (msglist[1] == "config") then
		if(RecipeBook_OptionsFrame:IsVisible()) then
			RecipeBook_OptionsFrame:Hide();
		else
			RecipeBook_OptionsFrame:Show();
		end
	--/rb output [chatframe|tooltip|both] [on|off] : Change output location
	elseif(msglist[1] == "output") then 
		if msglist[2] then 
			if msglist[3] then 
				if msglist[3] == "on" then toggle = "On";
				elseif msglist[3] == "off" then toggle = "Off";
				else RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
				end
				if msglist[2] == "chatframe" then 
					RecipeBook_SetOption("ChatFrame", toggle);
				elseif msglist[2] == "tooltip" then 
					RecipeBook_SetOption("Tooltip", toggle);
				elseif msglist[2] == "both" then
					RecipeBook_SetOption("ChatFrame", toggle);
					RecipeBook_SetOption("Tooltip", toggle);
				else 
					RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
				end
			else
				RecipeBook_Print(RECIPEBOOK_INVALID_OUTPUT, 1);
			end
		end
		RecipeBook_PrintStatus("Output");
	--/rb on : Turn RecipeBook on
	elseif(msg == "on") then 
		RecipeBook_SetOption("Status", "On");
		RecipeBook_PrintStatus("Status");
	--/rb off : Turn RecipeBook off
	elseif(msg == "off") then 
		RecipeBook_SetOption("Status", "Off");
		RecipeBook_PrintStatus("Status");
	--/rb help : Display usage
	elseif (msg == "help") then 
		RecipeBook_Print(RECIPEBOOK_USAGE, -1);
	--/rb status : Display RecipeBook on/off status
	elseif(msg == "status") then  RecipeBook_PrintStatus("Status")
	--/rb version : Display RecipeBook version number
	elseif(msg == "version") then RecipeBook_Print(RECIPEBOOK_VERSION, 0);
	--/rb skill [alt] [tradeskill] Print alt list for a particular tradeskill
	elseif (msglist[1] == "skill") then 
		if(msglist[2] and msglist[3]) then 
			msglist[3] = string.gsub(table.concat(msglist, " ", 3), "(%l)([%w_']*)", function(a,b)return string.upper(a)..string.lower(b)end);
			RecipeBook_Debug("Calling RB_SkillDisplay on "..msglist[2].." | "..msglist[3]);
			RecipeBook_SkillDisplay(msglist[2], msglist[3]);
		elseif not msglist[2] then 
			return RecipeBook_Print(RECIPEBOOK_NOALTGIVEN, 1);
		elseif not msglist[3] then 
			return RecipeBook_Print(RECIPEBOOK_NOTRADESKILLGIVEN, 1);
		end
	--/rb faction [same|opposite|both] : Change factional display
	elseif (msglist[1] == "faction") then
		if msglist[2] then
			if(msglist[2] == "same") then 
				RecipeBook_SetOption("SameFaction", "On");
				RecipeBook_SetOption("OtherFaction", "Off");
			elseif(msglist[2] == "opposite") then 
				RecipeBook_SetOption("SameFaction", "Off");
				RecipeBook_SetOption("OtherFaction", "On");
			elseif(msglist[2] == "both") then 
				RecipeBook_SetOption("SameFaction", "On");
				RecipeBook_SetOption("OtherFaction", "On");
			else 
				RecipeBook_Print(RECIPEBOOK_INVALID_DISPLAY, 1);
			end
		end
		RecipeBook_PrintStatus("AltFaction");
	--/rb debug : Debug toggle
	elseif(msglist[1] == "debug") then
		if (RB_Debug) then 
			RB_Debug = false;
		else 
			RB_Debug = true;
		end
		RecipeBook_Print("Debug: "..(RB_Debug and "ON" or "OFF"), 0);
	--/rb self [on|off] : Show/hide the current character
	elseif(msglist[1] == "self") then RecipeBook_ToggleOption("ShowSelf", msglist[2]);
	--/rb known [on|off] : Show/hide already known
	elseif(msglist[1] == "known") then
		if msglist[2] then RecipeBook_ToggleOption("Known", msglist[2])
		else RecipeBook_PrintStatus("Known");
		end
	--/rb learn [on|off] : Show/hide can be learned by
	elseif(msglist[1] == "learn") then
		if msglist[2] then RecipeBook_ToggleOption("CanLearn", msglist[2])
		else RecipeBook_PrintStatus("CanLearn");
		end
	--/rb future [on|off] : Show/hide will be learnable by
	elseif msglist[1] == "future" then
		if msglist[2] then RecipeBook_ToggleOption("WillLearn", msglist[2])
		else RecipeBook_PrintStatus("WillLearn");
		end
	--/rb banked : Show/hide banked status
	elseif msglist[1] == "banked" then
		if msglist[2] then RecipeBook_ToggleOption("Banked", msglist[2])
		else RecipeBook_PrintStatus("Banked");
		end
	--/rb specials : Scans for tradeskill specializations.
	elseif msglist[1] == "specials" then
		RecipeBook_SpecialtyScan();
	--/rb banklist : Lists all the items in the bank.
	elseif msglist[1] == "banklist" then
		RecipeBook_Print("Banked Items (for this faction only):", 0)
		table.foreach(RecipeBookData[Realm][Faction]["Banked"], function(a,b) RecipeBook_Print(b) end);
	--/rb bank <link> : Adds an item to the banked items list for that character.
	elseif msglist[1] == "bank" then
		local regex = ".*%[(.*)%].*";
		local item;
		if msglist[2] then 
			item = string.gsub(msg, regex, "%1") 
			if(item == msg) then return RecipeBook_Print(RECIPEBOOK_BANK_FAIL, 1) end;
			if(item and string.len(item) > 5) then
				table.insert(RecipeBookData[Realm][Faction]["Banked"], item);
				return RecipeBook_Print(format(RECIPEBOOK_BANK_SUCCEED, item), 0);
			end
		end
		return RecipeBook_Print(RECIPEBOOK_BANK_FAIL, 1); --no success.
	--/rb unbank <link> : Removes an item from the banked items list for that character.
	elseif msglist[1] =="unbank" then
		local regex = ".*%[(.*)%].*";
		local item;
		if msglist[2] then 
			item = string.gsub(msg, regex, "%1") 
			if(item == msg) then return RecipeBook_Print(RECIPEBOOK_UNBANK_FAIL, 1) end;
			local tempable = {};
			if(item and string.len(item) > 5) then
				table.foreach(RecipeBookData[Realm][Faction]["Banked"], function(a,b) if string.lower(b) ~= string.lower(item) then table.insert(tempable, b) end end);
				RecipeBookData[Realm][Faction]["Banked"] = tempable;
				return RecipeBook_Print(format(RECIPEBOOK_UNBANK_SUCCEED, item), 0);
			end
		end
		return RecipeBook_Print(RECIPEBOOK_UNBANK_FAIL, 1); --no success.
	--/rb reset <tradeskill> : Removes tradeskill data for the current character.
	elseif msglist[1] == "reset" then
		if msglist[2] then RecipeBook_ResetData(msglist[2])
		else RecipeBook_Print(RECIPEBOOK_INVALID_RESET, 1);
		end
	--/rb clear <character> : Removes all RecipeBook data for a particular character (not the current one).
	elseif msglist[1] == "clear" then
		if msglist[2] then
			who = string.gsub(msglist[2], "%l", string.upper, 1)
			if who == Player then return RecipeBook_Print(RECIPEBOOK_ERROR_RESETSELF, 1) end;
			if(RecipeBookData[Realm][Faction]["Personal"][who]) then --alt exists, same faction
				RecipeBookData[Realm][Faction]["Personal"][who] = nil;
				RecipeBook_Print(RECIPEBOOK_RESET_ALL..who, 0)
			elseif(RecipeBookData[Realm][Faction]["Shared"]) ~= nil then --shared data exists
				if RecipeBookData[Realm][Faction]["Shared"][who] ~= nil then --friend exists, same faction
					RecipeBookData[Realm][Faction]["Shared"][who] = nil;
					RecipeBook_Print(RECIPEBOOK_RESET_ALL..who, 0)
				end
			elseif(RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")] ~= nil) then --opposite faction exists
				if(RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Personal"] ~= nil) then --personal data exists
					if RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Personal"][who] ~= nil then --alt exists, opposite faction
						RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Personal"][who] = nil;
						RecipeBook_Print(RECIPEBOOK_RESET_ALL..who, 0)
					end
				elseif (RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Shared"] ~= nil) then --shared data exists
					if RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Shared"][who] ~= nil then --friend exists, opposite faction
						RecipeBookData[Realm][(Faction== "Alliance" and "Horde" or "Alliance")]["Shared"][who] = nil;
						RecipeBook_Print(RECIPEBOOK_RESET_ALL..who, 0)
					else --no such data exists
						RecipeBook_Print(RECIPEBOOK_NOALTMATCH..who, -1);
					end
				else
					RecipeBook_Print(RECIPEBOOK_NOALTMATCH..who, -1);
				end
			else
				RecipeBook_Print(RECIPEBOOK_NOALTMATCH..who, -1);
			end
		end
	--/rb send <player> <tradeskill>
	elseif(msglist[1] == "send") then
		if(msglist[2] and msglist[3]) then 
			RB_Messenger_InitiateSession(string.gsub(msglist[2], "^%l", string.upper), string.gsub(table.concat(msglist, " ", 3), "(%l)([%w_']*)", function(a,b)return string.upper(a)..string.lower(b)end))
		elseif (msglist[2] and msglist[2] == "kill") then RB_SessionActive = false; RecipeBook_Print("Session Killed.", 0);
 		else RecipeBook_Print(RECIPEBOOK_SEND_USAGE, 1)
		end
	--/rb receive <friend|guild|others> <accept|decline|prompt>
	elseif(msglist[1] == "receive") then
		if msglist[2] and msglist[3] then
			local op, set;
			if (string.find("friend guild others", msglist[2]) and string.find("accept decline prompt", msglist[3]))then 
				op = string.sub(msglist[2], 1, 1);
				set = string.upper(string.sub(msglist[3], 1, 1));
			else 
				RecipeBook_Print(RECIPEBOOK_RECEIVE_USAGE, 1);
				return;
			end
			if ((not IsInGuild()) and op == "g") then return RecipeBook_Print(RECIPEBOOK_NOTINGUILD, 1) end;
			local option = RecipeBook_GetOption("Receive")
			local regex = "("..op.."[ADP])";
			option = string.gsub(option, regex, op..set);
			RecipeBook_SetOption("Receive", option);
			RecipeBook_PrintStatus("Receive");
		elseif msglist[2] then
			RecipeBook_Print(RECIPEBOOK_RECEIVE_USAGE, 1);
		else
			RecipeBook_PrintStatus("Receive");	
		end
	--/rb friend <on|off> | faction<same|opposite|both> | known <on|off> | learn <on|off> | future <on|off> 
	elseif(msglist[1] == "friend") then
		--/rb friend on|off
		if ((msglist[2] == "on") or (msglist[2] == "off")) then 
			RecipeBook_SetOption("FriendShow", string.gsub(msglist[2], "^%l", string.upper));
			RecipeBook_PrintStatus("FriendShow");
		--/rb friend faction
		elseif msglist[2] == "faction" then 
			if msglist[3] then
				if(msglist[3] == "same") then 
					RecipeBook_SetOption("FriendSameFaction", "On");
					RecipeBook_SetOption("FriendOtherFaction", "Off");
				elseif(msglist[3] == "opposite") then 
					RecipeBook_SetOption("FriendSameFaction", "Off");
					RecipeBook_SetOption("FriendOtherFaction", "On");
				elseif(msglist[3] == "both") then 
					RecipeBook_SetOption("FriendSameFaction", "On");
					RecipeBook_SetOption("FriendOtherFaction", "On");
				else 
					RecipeBook_Print(RECIPEBOOK_INVALID_DISPLAY, 1);
				end
			end
			RecipeBook_PrintStatus("FriendAltFaction");
		--/rb friend known
		elseif msglist[2] == "known" then
			if msglist[3] then RecipeBook_ToggleOption("FriendKnown", msglist[3])
			else RecipeBook_PrintStatus("FriendKnown");
			end
		--/rb friend learn
		elseif msglist[2] == "learn" then
			if msglist[3] then RecipeBook_ToggleOption("FriendCanLearn", msglist[3]);
			else RecipeBook_PrintStatus("FriendCanLearn");
			end
		--/rb friend future 
		elseif msglist[2] == "future" then
			if msglist[3] then RecipeBook_ToggleOption("FriendWillLearn", msglist[3]);
			else RecipeBook_PrintStatus("FriendWillLearn");
			end
		end
	--No valid argument given.
	else 
		RecipeBook_Print(RECIPEBOOK_USAGE, -1);
	end
end

function RecipeBook_ToggleOption(option, value)
	if value == "on" then RecipeBook_SetOption(option, "On");
	elseif value == "off" then RecipeBook_SetOption(option, "Off");
	else RecipeBook_Print(RECIPEBOOK_INVALID_ONOFF, 1);
	end
	RecipeBook_PrintStatus(option);
end

--== [ HOOKED FUNCTIONS ] ==--
--[ Vanilla tooltip Updating functions ]--
function RecipeBook_ContainerFrameItemButton_OnEnter()
	RecipeBook_Old_ContainerFrameItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
function RecipeBook_AuctionFrameItem_OnEnter(type, index)
	RecipeBook_Old_AuctionFrameItem_OnEnter(type, index);
	RecipeBook_DoHookedFunction();
end
function RecipeBook_SetItemRef(link, name, button)
	RecipeBook_Old_SetItemRef(link, name, button);
	RB_TooltipUpdated = false;
	RecipeBook_DoHookedFunction("ItemRefTooltip");
end
--[ GameTooltip:SetInventoryItem() ]--
function RecipeBook_GameTooltip_SetInventoryItem(this, unit, slot) --Specifically a bank hack.
--	if slot < 40 then 
--		return RecipeBook_Old_GameTooltip_SetInventoryItem(this, unit, slot) 
--	end --Not a bank bag.  Die.
	local hasItem, hasCooldown, repairCost = RecipeBook_Old_GameTooltip_SetInventoryItem(this, unit, slot);
	RecipeBook_DoHookedFunction(); 
	return hasItem, hasCooldown, repairCost;
end
--[ GameTooltip:SetLootItem() ]---
function RecipeBook_GameTooltip_SetLootItem(this, slot)
	RecipeBook_Old_GameTooltip_SetLootItem(this, slot)
	RecipeBook_DoHookedFunction();
end
--[ GameTooltip:SetMerchantItem() ]--
function RecipeBook_GameTooltip_SetMerchantItem(unit, slot) --Vendors
	RecipeBook_Old_GameTooltip_SetMerchantItem(unit, slot);
	RecipeBook_DoHookedFunction(); 
end
--[ GameTooltip:OnHide() ]--
function RecipeBook_GameTooltip_OnHide()
	RB_TooltipUpdated = false;
	RB_ChatFrameUpdated = false;
	RecipeBook_Old_GameTooltip_OnHide();
end
--[ Container Frame ItemButton OnUpdate Hook ]--
function RecipeBook_ContainerFrameItemButton_OnUpdate(elapsed)
	RecipeBook_Old_ContainerFrameItemButton_OnUpdate(elapsed);
	RecipeBook_DoHookedFunction()
end

--==[ My Inventory ]==--
--[ MyInventory FrameItemButton OnEnter Hook ]--
function RecipeBook_MIFrameItemButton_OnEnter() --MyInventory doesn't call the original OnEnter function
	RecipeBook_Old_MIFrameItemButton_OnEnter();
	if (not RB_TooltipUpdated) then 
		RecipeBook_DoHookedFunction();
	end
end
--[ MyInventory FrameItemButton OnUpdate Hook ]--
function RecipeBook_MIFrameItemButton_OnUpdate(elapsed) --MyInventory doesn't call the original OnUpdate function
	RecipeBook_Old_MIFrameItemButton_OnUpdate(elapsed);
	RecipeBook_DoHookedFunction();
end
--==[ All-In-One-Inventory ]==--
--[ AIOI Modify Tooltip Hook ]--
function RecipeBook_AIOI_ModifyItemTooltip(bag, slot, tooltipName)
	RecipeBook_Old_AIOI_ModifyItemTooltip(bag, slot, tooltipName);
	RecipeBook_DoHookedFunction();
end
--==[ BankStatement ]==--
--[ BS ItemButton OnEnter Hook ]-- 
function RecipeBook_BS_ItemButton_OnEnter()
	RecipeBook_Old_BS_ItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
--[ BS Container ItemButton OnEnter Hook ]--
function RecipeBook_BS_ContainerFrameItemButton_OnEnter()
	RecipeBook_Old_BS_ContainerFrameItemButton_OnEnter();
	RecipeBook_DoHookedFunction();
end
--==[ BankItems ]==--
--[BI Button OnEnter Hook ]--
function RecipeBook_BI_Button_OnEnter()
	RecipeBook_Old_BI_Button_OnEnter();
	RecipeBook_DoHookedFunction();
end
--[ BI BagItem OnEnter Hook ]--
function RecipeBook_BI_BagItem_OnEnter()
	RecipeBook_Old_BI_BagItem_OnEnter();
	RecipeBook_DoHookedFunction();
end

--== [ DO THE HOOKED FUNCTION ] ==--
function RecipeBook_DoHookedFunction(tooltip)
	if RecipeBook_GetOption("Status") then 
		if not tooltip then tooltip = "GameTooltip" end;
		local text = getglobal(tooltip.."TextLeft1"):GetText();
		if text == nil then return end; --No top line text?
	
		if (RB_LastWhat[1] == text) and (not RB_LastWhat[2]) then 
			return; --Same name as last time, no need to mod tooltip.
		elseif (RB_LastWhat[1] ~= text) then --Different item, clear the modded data.
			RB_LastWhat[2] = false;
			RB_LastWhat[3] = {};
		end
		
		if tooltip == "GameTooltip" then 
			if not GameTooltip:IsVisible() then
				RB_LastWhat = {"RecipeBook: Blank Tooltip", false, {}}
				return;
			end
			
			if (RB_LastWhat[1] == text) and (RB_LastWhat[2]) then --Same name as last time, previously modded.
				RecipeBook_Debug("Scanning Tooltip.");
				local index;
				for index = GameTooltip:NumLines(), 4, -1 do
					local text = getglobal("GameTooltipTextLeft"..index):GetText();
					if text ~= nil and (string.find(text, RECIPEBOOK_KNOWNBY) or string.find(text, RECIPEBOOK_CANLEARN) or string.find(text, RECIPEBOOK_WILLLEARN) or string.find(text, RECIPEBOOK_ISBANKED)) then 
						return;
					end
				end
			end
		end;

		if text ~= nil then 
			local altsknow, altslearn, altshaveskill, isbanked, recipename, tradeskill, skillevel, specialty;
			if {RB_LastWhat[1] == text} and RB_LastWhat[2] and {RB_LastWhat[3] ~= {}} then
				altsknow = RB_LastWhat[3][1];
				altslearn = RB_LastWhat[3][2];
				altshaveskill = RB_LastWhat[3][3];
				isbanked = RB_LastWhat[3][4];
				recipename = string.gsub(text, "(%w+)%: (.+)", function(a,b) return b end);
			else
				RB_LastWhat[2] = false;
				RB_LastWhat[3] = {};
				recipename, tradeskill, skillevel, specialty = RecipeBook_GetRecipeData(tooltip, text);  --Let's start at the very beginning.
				if(recipename ~= nil) then 
					altsknow, altslearn, altshaveskill, isbanked = RecipeBook_MatchRecipeData(recipename, tradeskill, skillevel, specialty);
				else 
					RB_LastWhat = {text, false, {}};
					return;
				end
			end
			
			if recipename ~= nil then
				if RecipeBook_GetOption("ChatFrame") and (not RB_ChatFrameUpdated) and (RB_LastWhat[1] ~= text) then
					RB_LastWhat = {text, false, {altsknow, altslearn, altshaveskill, isbanked}};
					RecipeBook_OutputRecipeChatFrame(recipename, altsknow, altslearn, altshaveskill, isbanked);
				end
				if RecipeBook_GetOption("Tooltip") then
					RB_LastWhat = {text, false, {altsknow, altslearn, altshaveskill, isbanked}};
					RecipeBook_OutputRecipeTooltip(tooltip, altsknow, altslearn, altshaveskill, isbanked); --This will note if the tooltip was modified.
				end
			else
				RB_LastWhat = {text, false, {}}; 
			end
		end
	end
end



--== [ HANDLING TRADESKILL UPDATES ] ==--b
--[ UpdateSkillInfo : When the Tradeskill/Craft window opens, this wipes and replaces the known recipes list for that player. ]--
function RecipeBook_UpdateSkillInfo(skilltype)
	if (not Realm) or (not Faction) or (not Player) then return end;  --not yet initialized completely.
	local skillrank,maxranks,total;
	if(skilltype == "craft") then
		total = GetNumCrafts();
		skill, skillrank, maxranks = GetCraftDisplaySkillLine();
	else
		total = GetNumTradeSkills();
		skill, skillrank, maxranks = GetTradeSkillLine();
	end
	if skill == nil then 
		return false;
	end --Apparently hunter skills trigger the Tradeskill/Craft window too.
	local tradeskilltable = {};
	if(RecipeBookData[Realm][Faction]["Personal"][Player][skill]) then
		tradeskilltable.Specialty = RecipeBookData[Realm][Faction]["Personal"][Player][skill].Specialty;
		tradeskilltable.SubSpecialty = RecipeBookData[Realm][Faction]["Personal"][Player][skill].SubSpecialty;
	end
	
--	RecipeBookData[Realm][Faction]["Personal"][Player][skill] = {}; 
	tradeskilltable.Rank = tonumber(skillrank);
	local name,temp,crafttype;
	for index = 0, total, 1 do
		if(skilltype == "craft") then 
			name, temp, crafttype, temp, temp = GetCraftInfo(index);
		else 
			name, crafttype, temp, temp = GetTradeSkillInfo(index);
		end
		if((crafttype ~= "header") and (name ~= nil)) then
			table.insert(tradeskilltable, name);
		end
	end
	if(tradeskilltable ~= nil and RecipeBookData[Realm][Faction]["Personal"][Player][skill]~= nil) then
		if(table.getn(tradeskilltable) > table.getn(RecipeBookData[Realm][Faction]["Personal"][Player][skill])) then
			RecipeBookData[Realm][Faction]["Personal"][Player][skill] = tradeskilltable;
		else
			RecipeBookData[Realm][Faction]["Personal"][Player][skill].Rank = tonumber(skillrank);
		end
	elseif(tradeskilltable ~= nil) then 
		RecipeBookData[Realm][Faction]["Personal"][Player][skill] = tradeskilltable;
	end
end
--[ ParseSkillupMessage : When you get a skillup, update the tradeskill rank only. ]--
function RecipeBook_ParseSkillupMessage(text)
	local skill, rank;
	gsub(text, RECIPEBOOK_CHAT_SKILLUP, function(a,b) skill = a; rank = b; end);
	if(RecipeBookData[Realm][Faction]["Personal"][Player][skill] ~= nil) then 
		(RecipeBookData[Realm][Faction]["Personal"][Player][skill]).Rank = tonumber(rank);
	end
end

--[ ParseSystemMessage : Parses out learned recipe/spell names. Currently just eating CPU cycles while I figure out what to do with those names. ]--
function RecipeBook_ParseSystemMessage(text)
	local recipe, regex;
	if(string.find(text, RECIPEBOOK_CHAT_LEARN_RECIPE)) then
		regex = RECIPEBOOK_CHAT_LEARN_RECIPE.."([%w%s]*)%.";
	elseif(string.find(text, RECIPEBOOK_CHAT_LEARN_SPELL)) then
		regex = RECIPEBOOK_CHAT_LEARN_SPELL.."(%w)*%.";
	else
		return nil;
	end
	if(regex) then 
		gsub(text, regex, function(a) recipe = a end);
		return recipe;
	else
		return nil;
	end
end
--[ SpecialtyScan : /rb specials will scan your skills for tradeskill specializations. ]--
function RecipeBook_SpecialtyScan()
	local SpecString = "None";
	local SpecList = {};
	
	local name, texture, offset, numSpells = GetSpellTabInfo(1);
	for i = 1, numSpells, 1 do 
		local spellName, temp = GetSpellName(i, BOOKTYPE_SPELL);
		if(spellName) then
			RecipeBook_Debug("Matching Spell Name : "..spellName);
			local spellfind = "|"..spellName.." |";
			if string.find(RECIPEBOOK_BLACKSMITHING_SPECIALS, spellfind, 1, true) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_BLACKSMITHING] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_BLACKSMITHING] = {};
				end
				if(string.find(spellName, RECIPEBOOK_MASTER)) then 
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_BLACKSMITHING].Specialty = RECIPEBOOK_WEAPONSMITH;
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_BLACKSMITHING].SubSpecialty = spellName;
					table.insert(SpecList, RECIPEBOOK_WEAPONSMITH.." ("..spellName..")");
				else
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_BLACKSMITHING].Specialty = spellName;
					table.insert(SpecList, spellName);
				end
			end
			if string.find(RECIPEBOOK_ENGINEERING_SPECIALS, spellfind, 1, true) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_ENGINEERING] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_ENGINEERING] = {};
				end
				RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_ENGINEERING].Specialty = spellName;
				table.insert(SpecList, spellName);
			end
			if string.find(RECIPEBOOK_LEATHERWORKING_SPECIALS, spellfind, 1, true) then 
				if(RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_LEATHERWORKING] == nil) then 
					RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_LEATHERWORKING] = {};
				end
				RecipeBookData[Realm][Faction]["Personal"][Player][RECIPEBOOK_LEATHERWORKING].Specialty = spellName;
				table.insert(SpecList, spellName);
			end
		end
	end
	if table.getn(SpecList) > 0 then SpecString = table.concat(SpecList, ", ") end;
	RecipeBook_Print(RECIPEBOOK_ADDED_SPECIALS..SpecString, 0);
end


--== [ CHECK THE TOOLTIP DATA ] ==--
function RecipeBook_GetRecipeData(tooltip, text)
	local recipename, tradeskill, skillevel, specialty;
	-- Take tooltip line 1, look for "X: Y".  If X is in the list of pattern names, Y is the name of the pattern.
	if string.find(RECIPEBOOK_RECIPEPREFIXES, string.gsub(text, "(%w+)%: (.+)", function(a,b) recipename = b; return a end)) then -- It's a recipe!
		-- Take tooltip line 2, look for "Requires X (#)" and return X and # if found.  Otherwise this will return "text" in toto, meaning it's a malformed tradeskill requirement.
		text = getglobal(tooltip.."TextLeft2"):GetText();
		if(text ~= nil) then 
			string.gsub(text, RECIPEBOOK_REQUIRED_STRING.." (%w+) %((%d+)%)", function(a,b) tradeskill = a; skillevel = b end); 
		else 
			return nil;
		end -- This line shouldn't be empty, but just in case.
		if tradeskill == text then 
			return nil;
		end --No tradeskill requirement, no go.
		-- Take tooltip line 3, look for "Requires XY", where X is not "Level" and Y doesn't have to exist (i.e. "Requires Armorsmithing" only has X), and return XY.
		text = getglobal(tooltip.."TextLeft3"):GetText();
		if(text ~= nil) then 
			if(string.find(text, RECIPEBOOK_REQUIRED_STRING)) then 
				specialty = string.gsub(text, RECIPEBOOK_REQUIRED_STRING.."%s*(%w+)(.*)", function(b,c) if not (b == RECIPEBOOK_LEVEL_STRING) then return b..c else return false end end)
			else
				specialty = false;
			end
		else 
			specialty = false;
		end
		if text == specialty then 
			specialty =  false;
		end --No specialty requirement, that's cool.
		-- Put it all together, and what do you got?
		return recipename, tradeskill, tonumber(skillevel), specialty;
	else 
		return nil;  -- It's not a recipe!
	end
end


--== [ CHECK THE ALT DATA ] ==--
-- [ MatchRecipeData(recipename, tradeskill, rank, specialty) : Given tradeskill info, calls FindInFaction on each appropriate faction, then colour-codes appropriately. ] --
function RecipeBook_MatchRecipeData(recipename, tradeskill, rank, specialty)
	local knows, canlearn, hastradeskill, isbanked;
	local function tablestostring(table1, table2)
		local tstring = nil;
		table1 = table1 and table1 or {};
		table2 = table2 and table2 or {};
		if(table.getn(table1) > 0) then 
			tstring = table.concat(table1, ", ");
		end
		if(table.getn(table2) > 0) then
			if tstring then 
				tstring = tstring..", "..table.concat(table2, ", ");
			else 
				tstring = table.concat(table2, ", ");
			end
		end
		return tstring;
	end
	knows = {};
	canlearn = {};
	hastradeskill = {};
	if ((RecipeBook_GetOption("SameFaction")) and (RecipeBookData[Realm][Faction] ~= nil)) then 
		knows, canlearn, hastradeskill = RecipeBook_FindInFaction(Faction, "Personal", recipename, tradeskill, rank, specialty, knows, canlearn, hastradeskill);
	end
	if ((RecipeBook_GetOption("OtherFaction")) and (RecipeBookData[Realm][OtherFaction] ~= nil)) then 
		knows, canlearn, hastradeskill = RecipeBook_FindInFaction(OtherFaction, "Personal", recipename, tradeskill, rank, specialty, knows, canlearn, hastradeskill);
	end
	if(RecipeBook_GetOption("FriendShow")) then
		if ((RecipeBook_GetOption("FriendSameFaction")) and (RecipeBookData[Realm][Faction] ~= nil)) then 
			knows, canlearn, hastradeskill = RecipeBook_FindInFaction(Faction, "Shared", recipename, tradeskill, rank, specialty, knows, canlearn, hastradeskill);
		end
		if ((RecipeBook_GetOption("FriendOtherFaction")) and (RecipeBookData[Realm][OtherFaction] ~= nil)) then 
			knows, canlearn, hastradeskill = RecipeBook_FindInFaction(OtherFaction, "Shared", recipename, tradeskill, rank, specialty, knows, canlearn, hastradeskill);
		end
	end
	if(RecipeBook_GetOption("Banked")) then
		local function titletorecipe(text)
			return string.gsub(text, "(%w+)%: (.+)", function(c,d) return d end);
		end
		isbanked = table.foreach(RecipeBookData[Realm][Faction]["Banked"], function(a,b) if titletorecipe(b) == recipename then return true end end);
	end
	return table.concat(knows, ", "), table.concat(canlearn, ", "), table.concat(hastradeskill, ", "), isbanked;
end

-- [ FindInFaction(pass faction, recipename, tradeskill, rank, specialty) : Returns three lists of who knows, can learn, and has tradeskill for the passed faction. ] --
function RecipeBook_FindInFaction(lfaction, personal, recipename, tradeskill, rank, specialty, knows, canlearn, hastradeskill)
	local color;
	local doknown, docanlearn, dowilllearn;
	if(personal == "Personal") then 
		color = (lfaction == Faction) and RECIPEBOOK_FACTION_COLOR or RECIPEBOOK_OPPOSITE_COLOR;
		doknown = RecipeBook_GetOption("Known");
		docanlearn = RecipeBook_GetOption("CanLearn");
		dowilllearn = RecipeBook_GetOption("WillLearn");
	else
		color = (lfaction == Faction) and RECIPEBOOK_FRIEND_FACTION_COLOR or RECIPEBOOK_FRIEND_OPPOSITE_COLOR;
		doknown = RecipeBook_GetOption("FriendKnown");
		docanlearn = RecipeBook_GetOption("FriendCanLearn");
		dowilllearn = RecipeBook_GetOption("FriendWillLearn");
	end
	-- Matching recipes (see call).
	local function matchrecipe(who, skilltable)
		local prank, pspec, psubspec, pmatch;
		-- For each recipe in [Tradeskill][Recipe Table] if the recipe name matches then return true, else return nil.
		local function knowsrecipe (tableskill, recipetable)
			if tableskill == tradeskill then 
				pmatch = false;
				for index = 1, table.getn(recipetable), 1 do
					if(string.lower(string.gsub(string.gsub(recipetable[index], "[^%w]", ""),RECIPEBOOK_ENCHANTING_FILLER,"")) == string.lower(string.gsub(string.gsub(recipename, "[^%w]", ""),RECIPEBOOK_ENCHANTING_FILLER,""))) then 
						pmatch = true;
						break;
					end
				end
				prank = tonumber(recipetable.Rank);
				pspec = recipetable.Specialty;
				psubspec = recipetable.SubSpecialty;
				return true;
			else
				return nil;
			end
		end
		
		-- Skip the current alt
		if who == Player and not (RecipeBook_GetOption("ShowSelf")) then  
			return nil;
		end
		-- For each tradeskill in [Player]'s tradeskill table, if the tradeskill matches then return tradeskill rank, tradeskill specialty, and parse the list of known recipes - else nil.
		table.foreach(skilltable, knowsrecipe);
		-- If a tradeskill match is found:
		if(prank) then 
			who = color..who;
			-- If it's known, insert into the known list.
			if (pmatch and doknown) then 
					table.insert(knows, who..RECIPEBOOK_END_COLOR);
			-- If a specialty is required for this recipe, check to make sure specialty is known.  If it's not, then it's not learnable.
			elseif specialty then 
				if (pspec and pspec == specialty) or (psubspec and psubspec == specialty) then 
					if ((prank >= rank) and docanlearn) then 
						table.insert(canlearn, who..RECIPEBOOK_END_COLOR); 
					elseif dowilllearn then
						table.insert(hastradeskill, who..color.." ("..prank..")"..RECIPEBOOK_END_COLOR);
					end
				else
				end
			-- No specialty required; matches already listed, just checking to see if it's learnable now.
			else
				if ((prank >= rank) and docanlearn) then 
					table.insert(canlearn, who..RECIPEBOOK_END_COLOR); 
				elseif dowilllearn then
					table.insert(hastradeskill, who..color.." ("..prank..")"..RECIPEBOOK_END_COLOR);
				end
			end
		-- No tradeskill match was returned.
		else
			return nil;
		end
	end
	
	if(RecipeBookData[Realm] and RecipeBookData[Realm][lfaction] and RecipeBookData[Realm][lfaction][personal]) then 
		table.foreach(RecipeBookData[Realm][lfaction][personal], matchrecipe);
	end
	return knows, canlearn, hastradeskill;
end

--== [ TRADESKILL LISTING ] ==--
--[ SkillDisplay(who, what) : Sets up all the various skill display options ]--
function RecipeBook_SkillDisplay(who, what)
	who = string.gsub(string.lower(who), "%l", string.upper, 1);
	what = string.gsub(string.lower(what), "%l", string.upper, 1);
	RecipeBook_Debug("Who: "..who.." | What: "..what);
	if who == "All" then --/rb skill all <tradeskill>: Shows all alts of yours (not shared) who know a tradeskill, and their current skill.
		local matchedalts = {" ","*** Tradeskiller data for "..what.." ***",  "Your faction:", "--------------------"};
		local function matchtradeskill(alt,skillist)
			RecipeBook_Debug("Entered Matchtradeskill on skill "..what);
			if skillist[what] ~= nil then
				local ttext = alt.." ("..(skillist[what]).Rank..")";
				if((skillist[what]).Specialty) then 
					ttext = ttext..", "..(skillist[what]).Specialty;
					if (skillist[what]).SubSpecialty then ttext = ttext.." ["..(skillist[what]).SubSpecialty.."]" end;
				end
				table.insert(matchedalts, ttext);
			end
		end

		local tlength = table.getn(matchedalts);
		table.foreach(RecipeBookData[Realm][Faction]["Personal"], matchtradeskill);
		if table.getn(matchedalts) == tlength then table.insert(matchedalts, "No tradeskiller matches") end;
		if(RecipeBookData[Realm][(Faction=="Alliance" and "Horde" or "Alliance")] ~= nil) then
			table.insert(matchedalts, " ");
			table.insert(matchedalts, "Opposite faction:");
			table.insert(matchedalts, "--------------------");
			tlength = table.getn(matchedalts);
			table.foreach(RecipeBookData[Realm][(Faction=="Alliance" and "Horde" or "Alliance")]["Personal"], matchtradeskill);
			if table.getn(matchedalts) == tlength then table.insert(matchedalts, "No tradeskiller matches") end;
		end
		
		RecipeBook_Print(matchedalts, -1, RecipeBook_GetChatFrame());
		RecipeBook_Print("Tradeskiller information for "..what.." output to RecipeBook chat tab.", 0)
	
	elseif what == "All" then --/rb skill <character> all : Shows all tradeskills and ranks for any given character (shared or personal).
		local matchedalts = {" "};
		local function matchtradeskill(skillist, skilltable)
			if (skillist ~= nil) and (skillist ~= "Options") then
				local ttext = skillist.." ("..skilltable.Rank..")";
				if(skilltable.Specialty) then 
					ttext = ttext..", "..skilltable.Specialty;
					if skilltable.SubSpecialty then ttext = ttext.." ["..skilltable.SubSpecialty.."]" end;
				end
			table.insert(matchedalts, ttext);
			end
		end
		
		if(RecipeBookData[Realm][Faction]["Personal"][who] ~= nil) then
			table.insert(matchedalts, "*** Tradeskill data for "..who.." ("..Faction..") ***");
			table.foreach(RecipeBookData[Realm][Faction]["Personal"][who], matchtradeskill);
		elseif(RecipeBookData[Realm][Faction]["Shared"][who] ~= nil) then
			table.insert(matchedalts, "*** Tradeskill data for friend "..who.." ("..Faction..") ***");
			table.foreach(RecipeBookData[Realm][Faction]["Shared"][who], matchtradeskill);
		elseif(RecipeBookData[Realm][OtherFaction] ~= nil) then
			if(RecipeBookData[Realm][OtherFaction]["Personal"][who] ~= nil) then
				table.insert(matchedalts, "*** Tradeskill data for "..who.." ("..(Faction=="Alliance" and "Horde" or "Alliance")..") ***");
				table.foreach(RecipeBookData[Realm][OtherFaction]["Personal"][who], matchtradeskill);
			elseif(RecipeBookData[Realm][OtherFaction]["Shared"][who] ~= nil) then
				table.insert(matchedalts, "*** Tradeskill data for friend "..who.." ("..OtherFaction..") ***");
				table.foreach(RecipeBookData[Realm][OtherFaction]["Shared"][who], matchtradeskill);
			end
		end
		if table.getn(matchedalts) < 2 then return RecipeBook_Print("Cannot match alt name or no tradeskills for that alt.", 1) end;
		
		RecipeBook_Print(matchedalts, -1, RecipeBook_GetChatFrame());
		RecipeBook_Print("Tradeskill information for "..who.." output to RecipeBook chat tab.", 0)
	else--/rb skill <alt> <tradeskill>
		local alt, tradeskill, recipelist = RecipeBook_GetAltData(who, what);
		if alt then
			if(RB_SkillFrame:IsVisible()) then
				RB_TradeskillData = {};
				RB_SkillFrame:Hide();
			end
			RB_SkillFrame:Show();
			local specialty = "No specialty";
			if(recipelist.Specialty) then 
				specialty = recipelist.Specialty;
				if recipelist.SubSpecialty then specialty = specialty.." ["..recipelist.SubSpecialty.."]" end;
			end
			local rank = recipelist["Rank"] and recipelist["Rank"] or 0;
			RB_SkillFrameTitleText:SetText(alt);
			RB_SkillFrameTradeskillText:SetText(tradeskill.." ("..rank..")");
			RB_SkillFrameSpecialsText:SetText("Specialty: "..specialty);
			RB_TradeskillData = recipelist;
			RecipeBook_Debug("Table length (tradeskill data) before sort: "..table.getn(RB_TradeskillData));
			table.sort(RB_TradeskillData);
			RecipeBook_SkillScrollBar_Update();
		end
	end
end

--[ GetAltData(who, what) : Finds the appropriate tradeskill table for a given alt and tradeskill. ]--
function RecipeBook_GetAltData(who, what)
	local function FindAltTradeskill(skilltable)
		if (skilltable ~= nil) then
			return skilltable;
		else
			RecipeBook_Print(RECIPEBOOK_NOTRADESKILLMATCH..'"'..what..'".', 1);
			return false;
		end
	end
	
	local skilltable;
	if RecipeBookData[Realm][Faction]["Personal"][who] then
		skilltable = FindAltTradeskill(RecipeBookData[Realm][Faction]["Personal"][who][what]);
		if skilltable then return (who.." ("..Faction..")"), what, skilltable;
		else return false;
		end
	elseif (RecipeBookData[Realm][OtherFaction]) and (RecipeBookData[Realm][OtherFaction]["Personal"]) and (RecipeBookData[Realm][OtherFaction]["Personal"][who]) then
		skilltable = FindAltTradeskill(RecipeBookData[Realm][OtherFaction]["Personal"][who][what]);
		if skilltable then return (who.." ("..OtherFaction..")"), what, skilltable;
		else return false;
		end
	elseif RecipeBookData[Realm][Faction]["Shared"][who] then
		skilltable = FindAltTradeskill(RecipeBookData[Realm][Faction]["Shared"][who][what]);
		if skilltable then return (who.." (Friend, "..Faction..")"), what, skilltable;
		else return false;
		end
	elseif (RecipeBookData[Realm][OtherFaction]) and (RecipeBookData[Realm][OtherFaction]["Shared"]) and (RecipeBookData[Realm][OtherFaction]["Shared"][who]) then
		skilltable = FindAltTradeskill(RecipeBookData[Realm][OtherFaction]["Shared"][who][what]);
		if skilltable then return (who.." (Friend, "..OtherFaction..")"), what, skilltable;
		else return false;
		end
	else
		RecipeBook_Print(RECIPEBOOK_NOALTMATCH..'"'..who..'".', 1);
		return false;
	end
end
		
--== [ RESET A CHARACTER'S TRADESKILL DATA ] ==--
function RecipeBook_ResetData(tradeskill)
	tradeskill = string.gsub(tradeskill, "%l", string.upper, 1);
	if tradeskill  == "All" then
		local tempoptions = RecipeBook_GetOption("all");
		RecipeBookData[Realm][Faction]["Personal"][Player] = {};
		RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
		RecipeBook_UpdateOptions(tempoptions);
		RecipeBook_Print(RECIPEBOOK_RESET_ALL..Player, 0);
	else
		if RecipeBookData[Realm][Faction]["Personal"][Player][tradeskill] ~= nil then
			RecipeBookData[Realm][Faction]["Personal"][Player][tradeskill] = nil;
			RecipeBook_Print(RECIPEBOOK_RESET_SUCCEED..tradeskill, 0);
		else
			RecipeBook_Print(RECIPEBOOK_RESET_FAIL..tradeskill, 1);
		end
	end
end

--== [ OUTPUT TO TOOLTIP OR CHAT FRAME ] ==--
function RecipeBook_OutputRecipeTooltip(tooltip, knows, canlearn, havetradeskill, isbanked)
	local modded = false;
	if tooltip == "ItemRefTooltip" then 
		tooltip = ItemRefTooltip;
	else 
		tooltip = GameTooltip;
	end
	if(knows ~= "") then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_KNOWNBY..knows, 1,0,0);
		RB_LastWhat[2] = true;
		modded = true;
	end
	if(canlearn ~= "") then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_CANLEARN..canlearn, 0,1,0);
		RB_LastWhat[2] = true;
		modded = true;
	end
	if(havetradeskill ~= "") then
		if not modded then 
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_WILLLEARN..havetradeskill, 0,.75,.75);
		RB_LastWhat[2] = true;
		modded = true;
	end
	if(isbanked) then
		if not modded then
			tooltip:AddLine(" ",0,0,0);
		end
		tooltip:AddLine(RECIPEBOOK_ISBANKED, 0,1,0);
		RB_LastWhat[2] = true;
		modded = true;
	end 
	tooltip:Show();
	RB_TooltipUpdated = true;
end

function RecipeBook_OutputRecipeChatFrame(recipename, knows, canlearn, havetradeskill, isbanked)
	if RB_ChatFrameUpdated then return end;
	local chatframe = RecipeBook_GetChatFrame();
	if RecipeBook_GetOption("Known") then 
		if (knows ~= "") then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_KNOWNBY .. knows, -1, chatframe, 0, 1,0);
			RB_LastWhat[2] = true;
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONEKNOWN, -1, chatframe, 1,0,0);
		end
	end
	if RecipeBook_GetOption("CanLearn") then 
		if(canlearn ~= "") then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_CANLEARN .. canlearn, -1, chatframe, 0, 1,0);
			RB_LastWhat[2] = true;
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONELEARN, -1, chatframe, 1,0,0);
		end
	end
	if RecipeBook_GetOption("WillLearn") then
		if(havetradeskill ~= "") then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_WILLLEARN ..havetradeskill, -1, chatframe, 0, 1,0);
			RB_LastWhat[2] = true;
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NONEWILLLEARN, -1, chatframe, 1,0,0);
		end
	end
	if RecipeBook_GetOption("Banked") then
		if(isbanked) then
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_BANKED, -1, chatframe, 0, 1,0);
			RB_LastWhat[2] = true;
		else
			RecipeBook_Print(recipename .. RECIPEBOOK_CHATFRAME_NOTBANKED, -1, chatframe, 1,0,0);
		end
	end
	RB_ChatFrameUpdated = true;
end



--== [ UTILITY FUNCTIONS ] ==--
--[ Split : Splits a string into a table of words. ]--
function RecipeBook_Split(tocut)
	local words = {};
	function towords(word) table.insert(words, word) end
	if not string.find(string.gsub(tocut, "%S+", towords), "%S") then 
		return words;
	end
end

--[ Print(message, prefix (0 = <RecipeBook>, 1 = <RecipeBook Error:>, -1 = No Prefix), frame, colors) : Prints a RecipeBook status message. ]--
function RecipeBook_Print(message, prefix, frame, r,g,b)
	if message == nil then return end;
	if frame == nil then frame = DEFAULT_CHAT_FRAME end;
	if r == nil then r = 1; g = 1; b = 1;
	elseif g == nil then g = 1; b = 1;
	elseif b == nil then b = 1;
	end

	if(prefix == 0) then prefix = RECIPEBOOK_PREFIX;
	elseif(prefix == 1) then prefix = RECIPEBOOK_ERROR_PREFIX;
	else prefix = "";
	end
	if type(message) == "table" then
		table.foreach(message, function(t, u) frame:AddMessage(prefix..u, r, g, b) end )
	else
		frame:AddMessage(prefix..message, r, g, b);
	end
end
--[ Debug(message) : Prints a debug message if Debug is on. ]--
function RecipeBook_Debug(message)
	if(RB_Debug) then
		DEFAULT_CHAT_FRAME:AddMessage("<RecipeBook Debug:> "..message);
	end
end

--[ PrintStatus : Prints out the status of whatever options are passed to it. ]--
function RecipeBook_PrintStatus(option)
	toggleoptions = "Status FriendShow";
	showhideoptions = "Banked ShowSelf Known CanLearn WillLearn FriendKnown FriendCanLearn FriendShow FriendWillLearn";
	printorder = {"Status","Output","Receive","ShowSelf","FriendShow", "Banked", "AltFaction", "FriendAltFaction", "Known", "FriendKnown", "CanLearn", "FriendCanLearn", "WillLearn", "FriendWillLearn"};
	if(option == "all") then --Print all options
		RecipeBook_Print(RECIPEBOOOK_VERSION, 1);
		table.foreach(printorder, function(a,b) RecipeBook_PrintStatus(b) end);
	elseif(option == "Output") then --Print output status
		RecipeBook_Print(RECIPEBOOK_OUTPUT_TOOLTIP..(RecipeBook_GetOption("Tooltip") and RECIPEBOOK_ON or RECIPEBOOK_OFF), -1)
		RecipeBook_Print(RECIPEBOOK_OUTPUT_CHATFRAME..(RecipeBook_GetOption("ChatFrame") and RECIPEBOOK_ON or RECIPEBOOK_OFF), -1)
	elseif(option == "Receive") then
		local o = RecipeBook_GetOption(option);
		if (o == "fAgAoA") then RecipeBook_Print(RECIPEBOOK_ACCEPT_ALL);
		elseif(o == "fDgDoD") then RecipeBook_Print(RECIPEBOOK_DECLINE_ALL);
		elseif(o == "fPgPoP") then RecipeBook_Print(RECIPEBOOK_PROMPT_ALL);
		else
			local function FillOption(set)
				if set == "A" then return RECIPEBOOK_AUTOACCEPT;
				elseif set == "D" then return RECIPEBOOK_AUTODECLINE;
				else return RECIPEBOOK_PROMPT;
				end
			end
			RecipeBook_Print(RECIPEBOOK_RFRIENDS..FillOption(string.sub(o, 2, 2))..(IsInGuild() and (RECIPEBOOK_RGUILD..FillOption(string.sub(o, 4, 4))) or "")..RECIPEBOOK_ROTHERS..FillOption(string.sub(o, 6, 6)), -1);
		end
	elseif((option == "AltFaction") or (option == "FriendAltFaction")) then --Print factional display status
		if ((not RecipeBook_GetOption("FriendShow")) and option == "FriendAltFaction") then return end;
		local af = (option == "AltFaction" and "" or "Friend");
		if RecipeBook_GetOption(af.."SameFaction") then 
			if RecipeBook_GetOption(af.."OtherFaction") then
				RecipeBook_Print((af == "Friend" and RECIPEBOOK_FRIEND_PREFIX or "")..RECIPEBOOK_DISPLAY_BOTHFACTIONS, -1);
			else
				RecipeBook_Print((af == "Friend" and RECIPEBOOK_FRIEND_PREFIX or "")..string.format(RECIPEBOOK_DISPLAY_FACTION, Faction), -1);
			end
		elseif RecipeBook_GetOption(af.."OtherFaction") then
				RecipeBook_Print((af == "Friend" and RECIPEBOOK_FRIEND_PREFIX or "")..string.format(RECIPEBOOK_DISPLAY_FACTION, OtherFaction), -1);
		else
			RecipeBook_Print((af == "Friend" and RECIPEBOOK_FRIEND_PREFIX or "")..RECIPEBOOK_DISPLAY_NOFACTIONS, -1);
		end
	elseif string.find(toggleoptions, option) then
		RecipeBook_Print(string.format(getglobal("RECIPEBOOK_"..string.upper(option).."_ONOFF"), (RecipeBook_GetOption(option) and RECIPEBOOK_ON or RECIPEBOOK_OFF)), -1)
	elseif string.find(showhideoptions, option) then
		if(string.sub(option, 1, 6) == "Friend") then
			if RecipeBook_GetOption("FriendShow") then 
				RecipeBook_Print(RECIPEBOOK_FRIEND_PREFIX..string.format(getglobal("RECIPEBOOK_"..string.upper(string.sub(option, 7)).."_SHOWHIDE"), (RecipeBook_GetOption(option) and RECIPEBOOK_SHOW or RECIPEBOOK_HIDE)), -1)
			end
		else
			RecipeBook_Print(string.format(getglobal("RECIPEBOOK_"..string.upper(option).."_SHOWHIDE"), (RecipeBook_GetOption(option) and RECIPEBOOK_SHOW or RECIPEBOOK_HIDE)), -1)
		end
	end
end

--[ GetChatFrame (courtesy Averice's WIM, used by permission): Finds out which Chat Frame the RecipeBook info is in ]--
function RecipeBook_GetChatFrame()
	local frame = nil;
	local frametab, tempframe;
	for i = 1, NUM_CHAT_WINDOWS do
		tempframe = getglobal("ChatFrame" .. i .. "Tab");
		if (tempframe:GetText() == "Recipe Book") then
			frame = getglobal("ChatFrame" .. i);
			frametab = tempframe;
			break;
		end
	end
	if (frame) then
		if(not frametab:IsVisible()) then 
			frametab:Show(); 
		end
		return frame;
	else
		FCF_OpenNewWindow("Recipe Book");
		frame = RecipeBook_GetChatFrame(); 
		FCF_SetLocked(frame, true, false);
		ChatFrame_RemoveAllMessageGroups(frame);
		return frame;
	end
end

--== [ OPTIONS MODIFICATION FUNCTIONS ]==--
--[ UpdateOptions(options to set) : Updates the whole set of options at once. NOT idiot-proof.]--
function RecipeBook_UpdateOptions(newoptions)
	if(newoptions ~= nil) then 
		table.foreach(newoptions, RecipeBook_SetOption);
	end
end
--[ GetOption(option) : Returns the value of a given option; default if no value ]--
function RecipeBook_GetOption(option)
	if(option == "all") then 
		return RecipeBookData[Realm][Faction]["Personal"][Player]["Options"];
	elseif string.find(RECIPEBOOK_VALID_OPTIONS, option) then 
		if RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] ~= nil then
			option = RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option];
		else
			option = RECIPEBOOK_DEFAULT_OPTIONS[option];
		end
		if option == "On" then 
			return true;
		elseif option == "Off" then 
			return false;
		else 
			return option;
		end
	else
		return nil;
	end
end
--[ SetOption(option) : Sets the value of a given option.  Nil to reset to default. ]--
function RecipeBook_SetOption(option, value)
	if ((type(option) == "string") and (string.find(RECIPEBOOK_VALID_OPTIONS, option))) then
		if (value ~= nil) then 
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] = value;
		else
			RecipeBookData[Realm][Faction]["Personal"][Player]["Options"][option] = RECIPEBOOK_DEFAULT_OPTIONS[option];
		end
	end
end
--[ SetDefaultOptions() : Resets options, 'cause you may need it. ]--
function RecipeBook_SetDefaultOptions()
	RecipeBookData[Realm][Faction]["Personal"][Player]["Options"] = {};
end
----------------------------=== [ RecipeBook Config ] ==-----------------------------------
local CurrentOptions = {};

function RecipeBookConfig_OnShow()
	CurrentOptions = RecipeBook_GetOption("all");
end

function RecipeBookConfig_RevertOptions()
	CurrentOptions = RECIPEBOOK_DEFAULT_OPTIONS;
end

function RecipeBookConfig_OnHide()
end

function RecipeBookConfig_CommitOptions()
	RecipeBook_UpdateOptions(CurrentOptions);
end

function RecipeBookConfig_CheckboxClick(button, checked)
	CurrentOptions[button] = checked;
end

function RecipeBook_DD_FriendShare_OnLoad(frame)
	UIDropDownMenu_Initialize(this, RecipeBook_DD_FriendShare_Initialize);
	UIDropDownMenu_SetButtonWidth(60);
	UIDropDownMenu_SetWidth(80);
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_FriendShare, RecipeBook_OptionsFrame_GetReceiveOption("friend"));
end
function RecipeBook_DD_FriendShare_Initialize()
	local info;
	for i=1, table.getn(RECIPEBOOK_CONFIG_DD_SHARE), 1 do
		info = {};
		info.text = RECIPEBOOK_CONFIG_DD_SHARE[i].text;
		info.func = RecipeBook_DD_FriendShare_OnClick;
		UIDropDownMenu_AddButton(info);
	end
end
function RecipeBook_DD_FriendShare_OnClick()
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_FriendShare, this:GetID());
	CurrentOptions["Receive"] = string.gsub(RecipeBook_GetOption("Receive"), "(f%u)", ("f"..RECIPEBOOK_CONFIG_DD_SHARE[this:GetID()].name));
end

function RecipeBook_DD_GuildShare_OnLoad(frame)
	UIDropDownMenu_Initialize(this, RecipeBook_DD_GuildShare_Initialize);
	UIDropDownMenu_SetButtonWidth(60);
	UIDropDownMenu_SetWidth(80);
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_GuildShare, RecipeBook_OptionsFrame_GetReceiveOption("guild"));
	if not IsInGuild() then OptionsFrame_DisableDropDown(RecipeBook_DD_GuildShare) end;
end
function RecipeBook_DD_GuildShare_Initialize()
	local info;
	for i=1, table.getn(RECIPEBOOK_CONFIG_DD_SHARE), 1 do
		info = {};
		info.text = RECIPEBOOK_CONFIG_DD_SHARE[i].text;
		info.func = RecipeBook_DD_GuildShare_OnClick;
		UIDropDownMenu_AddButton(info);
	end
end
function RecipeBook_DD_GuildShare_OnClick()
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_GuildShare, this:GetID());
	CurrentOptions["Receive"] = string.gsub(RecipeBook_GetOption("Receive"), "(g%u)", ("g"..RECIPEBOOK_CONFIG_DD_SHARE[this:GetID()].name));
end

function RecipeBook_DD_OtherShare_OnLoad(frame)
	UIDropDownMenu_Initialize(this, RecipeBook_DD_OtherShare_Initialize);
	UIDropDownMenu_SetButtonWidth(60);
	UIDropDownMenu_SetWidth(80);
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_OtherShare, RecipeBook_OptionsFrame_GetReceiveOption("other"));
end
function RecipeBook_DD_OtherShare_Initialize()
	local info;
	for i=1, table.getn(RECIPEBOOK_CONFIG_DD_SHARE), 1 do
		info = {};
		info.text = RECIPEBOOK_CONFIG_DD_SHARE[i].text;
		info.func = RecipeBook_DD_OtherShare_OnClick;
		UIDropDownMenu_AddButton(info);
	end
end
function RecipeBook_DD_OtherShare_OnClick()
	UIDropDownMenu_SetSelectedID(RecipeBook_DD_OtherShare, this:GetID());
	CurrentOptions["Receive"] = string.gsub(RecipeBook_GetOption("Receive"), "(o%u)", ("o"..RECIPEBOOK_CONFIG_DD_SHARE[this:GetID()].name));
end

function RecipeBook_OptionsFrame_GetReceiveOption(id)
	if id == "friend" then id = 2;
	elseif id == "guild" then id = 4;
	else id = 6;
	end
	local curroption = string.sub(RecipeBook_GetOption("Receive"), id, id);
	if curroption == "A" then return 1; --curroption = RECIPEBOOK_CONFIG_TEXT_SHARE[1];
	elseif curroption == "D" then return 2; --curroption = RECIPEBOOK_CONFIG_TEXT_SHARE[2];
	else return 3; --curroption = RECIPEBOOK_CONFIG_TEXT_SHARE[3];
	end
end


--------------------------== [ RecipeBook Messenger ]==------------------------

--[ ChatFrame_OnEvent(event) : Catches whispers for RB communication. ]--
function RecipeBook_ChatFrame_OnEvent(event)
	local who, what;
	if((event == "CHAT_MSG_WHISPER") or (event == "CHAT_MSG_WHISPER_INFORM")) then
		what = arg1;
		who = arg2;
		if(string.find(what, RECIPEBOOK_MESSAGE_TRIGGER, 1, true)) then 
			RB_Messenger_ProcessMessage(who,what,event);
		else
			RecipeBook_Old_ChatFrame_OnEvent(event);
		end
	else
		RecipeBook_Old_ChatFrame_OnEvent(event);
	end
end

--[ ProcessMessage : Parses a sent/received message. ]--
local lastwhat = "";
function RB_Messenger_ProcessMessage(who,what,event)
	local whatskill;
	local start,finish;
	if(event == "CHAT_MSG_WHISPER") then
		if(lastwhat ~= what) then
			RecipeBook_Debug("Data catch: <"..what.."> from "..who);
			lastwhat = what;
			-- Receiving Initiate request
			if (string.find(what, RECIPEBOOK_MESSAGE_INITIATE)) then 
				if not RB_SessionActive then
					local regex = '.+' .. RECIPEBOOK_MESSAGE_INITIATE .. '(%w+)';
					whatskill = string.gsub(what, regex, "%1");
					RB_Messenger_ConditionalAcceptance(who, whatskill, RecipeBook_GetOption("Receive"));
				else
					SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_BUSY, WHISPER, this.language, who);
				end
			-- Receiving Initiate accept
			elseif string.find(what, RECIPEBOOK_MESSAGE_ACCEPT) and RB_Messenger_GetName() == who then
				StaticPopup_Hide("RECIPEBOOK_AWAITING_SEND")
				RB_Messenger_SendData(who, RB_SessionActive[2]);
			-- Receiving Busy keystring
			elseif string.find(what, RECIPEBOOK_MESSAGE_BUSY) then 
				RecipeBook_Print(format(RECIPEBOOK_ERROR_BUSY, who), 1);
				StaticPopup_Hide("RECIPEBOOK_AWAITING_SEND");
				RB_SessionActive = false;
			-- Receiving Terminate keystring
			elseif string.find(what, RECIPEBOOK_MESSAGE_TERMINATE) and RB_Messenger_GetName() == who then
				if(RecipeBookData[Realm][Faction]["Shared"][who] ~= nil) then 
					RecipeBookData[Realm][Faction]["Shared"][who][RB_SessionActive[2]] = RB_ReceivedData;
				else
					RecipeBookData[Realm][Faction]["Shared"][who] = {};
					RecipeBookData[Realm][Faction]["Shared"][who][RB_SessionActive[2]] = RB_ReceivedData;
				end
				RecipeBook_Print(RECIPEBOOK_ERROR_RECEIVED, 0);
				SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_ACKNOWLEDGE, WHISPER, this.language, who);
				RB_SessionActive = false;
			--Receiving Autodecline keystring
			elseif string.find(what, RECIPEBOOK_ERROR_AUTODECLINE) and RB_Messenger_GetName() == who then
					local regex = '.+' .. RECIPEBOOK_ERROR_AUTODECLINE .. '(%w+)';
					whatskill = string.gsub(what, regex, "%1");
					RecipeBook_Print(who..RECIPEBOOK_ERROR_AUTODECLINE..whatskill, 1);
				StaticPopup_Hide("RECIPEBOOK_AWAITING_SEND"); 
				RB_SessionActive = false;
			--Receiving Cancel keystring
			elseif string.find(what, RECIPEBOOK_MESSAGE_CANCEL) and RB_Messenger_GetName() == who then
				RecipeBook_Print(format(RECIPEBOOK_ERROR_CANCEL, who), 1);
				if (RB_SessionActive[1] == "Send"  or RB_SessionActive[1] == "Queue") then StaticPopup_Hide("RECIPEBOOK_AWAITING_SEND"); 
				else StaticPopup_Hide("RECIPEBOOK_REQUESTING_SEND");
				end
				RB_SessionActive = false;
			--Do nothing on an acknowledge receipt
			elseif string.find(what, RECIPEBOOK_MESSAGE_ACKNOWLEDGE) then 
			--Anything else is potential data.
			elseif RB_SessionActive and RB_SessionActive[1] == "Receive" and RB_Messenger_GetName() == who then
				RB_Messenger_ProcessData(what);
			end
		end
	elseif(event == "CHAT_MSG_WHISPER_INFORM") then
		if(lastwhat ~= what) then
			RecipeBook_Debug("Data send: <"..what.."> to "..who);
			lastwhat = what;
		end
	end
end

--[ InitiateSession (sendto, tradeskill to send) : Initiates RecipeBook sharing.  Aww, how nice. ]--
function RB_Messenger_InitiateSession(who,whatskill)
	if who == Player then return RecipeBook_Print(RECIPEBOOK_MATCHEDSELF, 1) end;
	if whatskill == "All" then
		RB_SessionQueue = {};
		local function recursivesend(skill,temp)
			if skill ~= "Options" then table.insert(RB_SessionQueue, skill) end;
		end
		table.foreach(RecipeBookData[Realm][Faction]["Personal"][Player], recursivesend);
		if table.getn(RB_SessionQueue) > 0 then 
			RB_Messenger_InitiateSession(who, "Queue");
			StaticPopup_Show("RECIPEBOOK_AWAITING_SEND");
		end
	elseif whatskill == "Queue" then
		whatskill = RB_SessionQueue[1];
		SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_INITIATE..whatskill, WHISPER, this.language, who);
		RecipeBook_Print(format(RECIPEBOOK_ERROR_INITIATE, whatskill, who), 0);
		table.remove(RB_SessionQueue, 1);
		if table.getn(RB_SessionQueue) < 1 then
			RB_SessionActive = {"Send", whatskill, who}
		else
			RB_SessionActive = {"Queue", whatskill, who};
		end
		--StaticPopup_Show("RECIPEBOOK_AWAITING_SEND");
	elseif(RecipeBookData[Realm][Faction]["Personal"][Player][whatskill] ~= nil) then 
		SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_INITIATE..whatskill, WHISPER, this.language, who);
		RecipeBook_Print(format(RECIPEBOOK_ERROR_INITIATE, whatskill, who), 0);
		RB_SessionActive = {"Send", whatskill, who};
		StaticPopup_Show("RECIPEBOOK_AWAITING_SEND");
	else
		return RecipeBook_Print(RECIPEBOOK_NOTRADESKILLMATCH..whatskill, 1);
	end
end

--[ AcceptSession (sendfrom, tradeskill) : Accepts the offered RB share ]--
function RB_Messenger_AcceptSession(who, whatskill)
	SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_ACCEPT..whatskill, WHISPER, this.language, who);
	RB_ReceivedData = {};
	RB_SessionActive = {"Receive", whatskill, who};
end

--[ ConditionalAcceptance(sendfrom, tradeskill, accept options) : Accepts, declines, or prompts for acceptance depending on who's sending the data ]--
function RB_Messenger_ConditionalAcceptance(who, whatskill, option)
	local function RB_AcceptByOption(o)
		if o == "A" then 
			RecipeBook_Print(string.format(RECIPEBOOK_ERROR_ACCEPTEDSESSION, who, whatskill), 0);
			RB_Messenger_AcceptSession(who, whatskill);
		elseif o == "D" then 
			RB_Messenger_CancelSession(who, whatskill, "decline");
		else
			RB_SessionActive = {"Pending", whatskill, who};
			StaticPopup_Show("RECIPEBOOK_REQUESTING_SEND");
		end
	end
	if option == "fPgPoP" then
		RB_AcceptByOption("P");
	elseif option == "fAgAoA" then
		RecipeBook_Print(string.format(RECIPEBOOK_ERROR_ACCEPTDEDSESSION, who, whatskill), 0);
		RB_Messenger_AcceptSession(who, whatskill);
	elseif option == "fDgDoD" then
		RB_Messenger_CancelSession(who, whatskill, "decline");
	else
		--friends first:
		local n;
		local friend = false;
		for n = 1, GetNumFriends() do
			if GetFriendInfo(n) == who then 
				friend = true;
				break;
			end
		end
		if friend then  --friend?
			RB_AcceptByOption(string.sub(option, 2,2));
		elseif (IsInGuild()) then--Guild member?
			if (GetNumGuildMembers() == 0) then GuildRoster(); end
			friend = false;
			for n = 1, GetNumGuildMembers() do
				if GetGuildRosterInfo(n) == who then
					friend = true;
					break;
				end
			end
			if friend then
				RB_AcceptByOption(string.sub(option, 4,4));
			end
		else --All other persons
				RB_AcceptByOption(string.sub(option, 6,6));
		end
	end
end

--[ CancelSession : Sends a terminate request that does NOT write data ]--
function RB_Messenger_CancelSession(who, whatskill, cancel)
	if(cancel == "decline") then
		SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..Player..RECIPEBOOK_ERROR_AUTODECLINE..whatskill, WHISPER, this.language, who);
		RecipeBook_Print(RECIPEBOOK_ERROR_DECLINEDSESSION..who, 0)
	else
		SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_CANCEL..whatskill, WHISPER, this.language, who);
		RecipeBook_Print(RECIPEBOOK_ERROR_CANCEL, 1);
	end
	RB_SessionActive = false;
end

--[ CancelSession : Sends a terminate request that initiates a data write ]--
function RB_Messenger_TerminateSession(who, whatskill)
	SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER..RECIPEBOOK_MESSAGE_TERMINATE..whatskill, WHISPER, this.language, who);
	if RB_SessionActive[1] == "Queue" then
		StaticPopup_Show("RECIPEBOOK_AUTO_SEND");
	else
		RB_SessionActive = false;
	end
	RecipeBook_CountdownFrame.listinuse = {{}, "nobody", "noskill"};
	RecipeBook_Print(RECIPEBOOK_ERROR_TERMINATE, 0);
end

--[ GetName : Returns the name of the player on the other end of the session ]--
function RB_Messenger_GetName()
	if(RB_SessionActive) then return RB_SessionActive[3] else return "No Player" end;
end

--[ GetTradeskill : Returns the name of the tradeskill currently being shared ]--
function RB_Messenger_GetTradeskill()
	if(RB_SessionActive) then return RB_SessionActive[2] else return "No Tradeskill" end;
end

--[ SendData : Sends tradeskill data for a given tradeskill ]--
function RB_Messenger_SendData(who, whatskill)
	local skill = RB_Messenger_TradeskillList(whatskill);
	local recipelist = skill[1];
	local skillspecs = skill[3];
	SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER.." Rank: "..skill[2], WHISPER, this.language, who);
	if skillspecs[1] ~= nil then
		SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER.." Specialty: "..skillspecs[1], WHISPER, this.language, who);
		if skillspecs[2] then
			SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER.." subSpecialty: "..skillspecs[2], WHISPER, this.language, who);
		end
	end
	RecipeBook_Debug("Beginning batch data send, length: "..table.getn(recipelist));
	RecipeBook_SendDataBatch(recipelist, who, whatskill, 0);
end

function RecipeBook_SendDataBatch(recipelist, who, whatskill, offset)
	local length = table.getn(recipelist) - offset;
	RecipeBook_Debug("Batch length: "..length);
	if length == 0 then 
		RB_Messenger_TerminateSession(who, whatskill);
	elseif length < 21 then
		RecipeBook_Debug("Beginning final data batch share");
		for index=1,length, 1 do
			if(recipelist[index] ~= nil) then 
				SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER.." : "..recipelist[index], WHISPER, this.language, who) 
			end
		end
		RecipeBook_SendDataBatch({}, who, whatskill, 0);
	else
		RecipeBook_Debug("Sending 20 data items.");
		for i = 1+offset, 20+offset, 1 do
			SendChatMessage(RECIPEBOOK_MESSAGE_TRIGGER.." : "..recipelist[i], WHISPER, this.language, who) 
		end
		offset = offset + 20;
		RecipeBook_Debug("Pausing.");
		RecipeBook_CountdownFrame:Show();
		RecipeBook_CountdownFrame.listinuse = {recipelist, who, whatskill, offset};
	end
end
	


--[ ProcessData : When RecipeBook data is received, parses out Rank/Specialty information and then stores recipes. ]--
function RB_Messenger_ProcessData(what)
	local regex = '%[' ..RECIPEBOOK_MESSAGE_TRIGGER_WORD .. '%] ' .. '(.+)';
	local whatstring = string.gsub(what, regex, "%1");
	if(string.sub(whatstring, 1, 4) == "Rank") then
		RB_ReceivedData.Rank = tonumber(string.sub(whatstring, 7));
	elseif(string.sub(whatstring, 1, 9) == "Specialty") then
		RB_ReceivedData.Specialty = string.sub(whatstring, 12);
	elseif(string.sub(whatstring, 1, 12) == "subSpecialty") then
		RB_ReceivedData.subSpecialty = string.sub(whatstring, 15);
	else
		table.insert(RB_ReceivedData, string.sub(whatstring, 3));
	end
end

--[ TradeskillList : Collects the recipe list, specialties, and ranks for the player and returns in a format RecipeBook can send.  Possibly a little redundant. ]--
function RB_Messenger_TradeskillList(what)
	if (what ~= "") then
		if(RecipeBookData[Realm][Faction]["Personal"][Player][what] ~= nil) then 
			recipelist = RecipeBookData[Realm][Faction]["Personal"][Player][what];
			RecipeBook_Debug(table.concat(recipelist, " | "));
			local skillrank = tonumber(recipelist.Rank);
			if(skillrank == nil) then skillrank = 0 end;
			local skillspecs = {};
			if recipelist.Specialty then
				skillspecs = {recipelist.Specialty, (recipelist.subSpecialty and recipelist.subSpecialty or nil)};
			end
			if (recipelist ~= nil)  then return {recipelist, skillrank, skillspecs} else return {} end;
		else return RecipeBook_Print(RECIPEBOOK_NOTRADESKILLMATCH..whatskill, 1);
		end
	else return RecipeBook_Print(RECIPEBOOK_NOTRADESKILLGIVEN, 1);
	end
end

-------------------==[ SKILL FRAME DISPLAY ]==----------------------
function RecipeBook_SkillScrollBar_Update()
	local line; 
	local lineplusoffset;
	local maxlength = table.getn(RB_TradeskillData) +1;
	RecipeBook_Debug("Tradeskill Data loaded, length: "..table.getn(RB_TradeskillData));
	FauxScrollFrame_Update(RecipeBook_Skill_ScrollFrame,maxlength,14,20);
	for line=1,14 do
		lineplusoffset = line + FauxScrollFrame_GetOffset(RecipeBook_Skill_ScrollFrame);
		RecipeBook_Debug("lineplusoffset: "..lineplusoffset);
		if lineplusoffset < maxlength then
			RecipeBook_Debug("Setting Text to: "..RB_TradeskillData[lineplusoffset]);
			getglobal("RecipeBook_SkillFrameEntry"..line.."_Text"):SetText(RB_TradeskillData[lineplusoffset]);
			getglobal("RecipeBook_SkillFrameEntry"..line):Show();
		else
			getglobal("RecipeBook_SkillFrameEntry"..line):Hide();
		end
	end
end

function RB_SkillFrame_OnShow()
	RB_SkillListFrame:Show();
end

function RB_SkillFrame_OnHide()
	RB_SkillListFrame:Hide();
end
 
function RecipeBook_CountdownUpdate(elapsed)
	if this.timeSinceLastUpdate == nil then this.timeSinceLastUpdate = 0 end;
	this.timeSinceLastUpdate = this.timeSinceLastUpdate + elapsed; 	
	if (this.timeSinceLastUpdate > RECIPEBOOK_SEND_PAUSE) then    
		this.timeSinceLastUpdate = 0;
		this:Hide();
	end
end
